import numpy as np
import os
from typing import Union

from .direction import Direction
from .grid_base import GridBase
from .grid_info import GridInfo
from .actions import Actions


class GridLevel():
  '''
    Class to manage a grid level.
    This performs all drawing and querying of the level
    (to see which moves are possible within a given grid cell)
  '''

  def __init__( self, **kwargs: dict ):

    # get the directory where this file is running
    dir_path = os.path.dirname(os.path.realpath(__file__))

    self.grid_base = GridBase( dir_path, **kwargs )
    self.grid_info = GridInfo( self.grid_base, **kwargs )

  '''
      Query Functions
  '''

  def is_grid_state( self, x, y ):
    ''' test if the specified state is a valid state
        i.e. it lies on the grid level
    '''
    if (x < 0) or (x >= self.grid_base.width) or \
       (y < 0) or (y >= self.grid_base.height) or \
       self.grid_base.test_for_base_area( x, y ):
      return False

    # the terminal state doesn't count as a valid grid state
    if (x == self.grid_base.end[0]) and (y== self.grid_base.end[1]):
      return False

    return True



  def get_directions( self, x: int = None, y: int = None ) -> Union[Direction,np.ndarray]:
    ''' return the possible directions for the specified grid cell
        - if a grid cell is not specified then return an array of possible
        directions for all grid cells
    '''
    return self.grid_info.get_directions( x, y )


  def get_rewards( self, x: int = None, y: int = None ) -> Union[Direction,np.ndarray]:
    ''' return the reward for the specified grid cell
        - if a grid cell is not specified then return an array of rewards for all grid cells
    '''
    return self.grid_base.get_reward( x, y )


  def get_action_probabilities( self, x, y, action: Actions ):

    direction = Direction.from_action( action )
    assert direction >= Direction.Stay and direction <= Direction.West

    # check that some actions are possible in this state
    possible_directions = self.grid_info.get_cell_directions(x,y,direction)

    if not possible_directions or direction==Direction.Stay:
      # stay in same position, reward is the same as if trying to move into this state
      # - target only reached if choosing to stay in same state
      reward = self.grid_base.get_reward( x, y )
      return [[1.0,[x,y],reward]]

    # get the probability of moving to the intended target
    transition_probability, barrier = self.grid_base.get_transition_probability( x, y, direction )

    action_list = []

    # find the details of where we end up if action succeeds
    if transition_probability > 0.0:
      next_pos = self.get_next_state_position( x, y, direction )
      reward = self.grid_base.get_reward( next_pos[0], next_pos[1] )
      action_list.append([transition_probability,next_pos,reward])

    # if the transition probability is 1 then definitely reach target
    if transition_probability == 1.0:
      return action_list

    #
    # transition_probability < 1.0
    # - more than one next_state for current action
    #

    # get the list of all other possible states
    all_actions = self.grid_info.get_cell_directions(x,y)
    all_actions.pop(direction.get_direction_char(), None)
    other_directions = [key for (key, value) in all_actions.items() if value]

    # if there are no other states to go to, stay in the current state
    if len(other_directions) == 0:
      reward = self.grid_base.get_reward( x, y )
      action_list.append([(1.0 - transition_probability),[x,y],reward])
      return action_list

    if barrier:
      # an extra penalty of -1 is given for running into a wall
      reward = -1

      # if a barrier existed in the target direction then Baby Robot will end up taking
      # the opposite action if this exists (i.e. he bounces off the barrier)
      # - this happens with probability (1-transition_probability)
      direction = Direction.get_opposite(direction).get_direction_char()

      # check that the opposite direction is possible
      if direction not in other_directions:
        # opposite direction not possible - stay in current state
        reward += self.grid_base.get_reward( x, y )
        action_list.append([(1.0 - transition_probability),next_pos,reward])
        return action_list

      next_pos = self.get_next_state_position( x, y, direction )
      reward += self.grid_base.get_reward( next_pos[0], next_pos[1] )
      action_list.append([(1.0 - transition_probability),next_pos,reward])
      return action_list

    # divide the remaining probability over the other possible directions
    probability = (1.0-transition_probability)/len(other_directions)
    for other_direction in other_directions:
      # calculate the postion of the next state
      next_pos = self.get_next_state_position( x, y, other_direction )
      reward = self.grid_base.get_reward( next_pos[0], next_pos[1] )
      action_list.append([probability,next_pos,reward])

    return action_list


  def get_next_state( self, x, y, direction ):
    ''' return the next state and reward for moving
        - (x,y) = current state
        - direction = direction moved from current state
    '''

    assert direction >= Direction.Stay and direction <= Direction.West

    # check that some actions are possible in this state
    possible_actions = self.grid_info.get_cell_directions(x,y,direction)
    if not possible_actions or direction==Direction.Stay:
      # stay in same position, reward is the same as if trying to move into this state
      # - target only reached if choosing to stay in same state
      reward = self.grid_base.get_reward( x, y )
      return [x,y],reward,(direction==Direction.Stay)

    # a deterministic policy should only have one possible action
    chosen_action = [key for (key, value) in possible_actions.items() if value]
    if len(chosen_action) != 1:
      # stay in same position, reward is the same as if trying to move into this state
      reward = self.grid_base.get_reward( x, y )
      return [x,y],reward,False

    # get the list of all other possible states
    all_actions = self.grid_info.get_cell_directions(x,y)
    all_actions.pop(chosen_action[0], None)
    other_states = [key for (key, value) in all_actions.items() if value]

    # get the probability of moving to the intended target
    transition_probability, barrier = self.grid_base.get_transition_probability( x, y, chosen_action[0] )

    # if the probability is less than the transition probability then move to the target
    # of if the target state is the only allowed state
    target_state_reached = True
    # if (np.random.random() < transition_probability) or (len(other_states) == 0):
    if (np.random.random() < transition_probability):
      direction = chosen_action[0]
    else:
      # the action did not succeed

      # set the flag to indicate the target state wasn't reached
      target_state_reached = False

      # if there are no other states to go to, stay in the current state
      if len(other_states) == 0:
        direction = Direction.Stay
      else:
        if barrier:
          # if a barrier existed in the target direction then Baby Robot will end up taking
          # the opposite action if this exists (i.e. he bounces off the barrier)
          direction = Direction.get_opposite(chosen_action[0])

          # check that the opposite direction is possible
          if direction not in other_states:
            return [x,y],-1,False

        else:
          # choose one of the other possible states
          direction = np.random.choice(other_states)

    # calculate the postion of the next state
    next_pos = self.get_next_state_position( x, y, direction )

    # get the reward for taking this action
    reward = self.grid_base.get_reward( next_pos[0], next_pos[1] )

    # if we bounced off a barrier incur an extra time penalty of -1
    if barrier:
      reward += -1

    # for equal probability of taking an action its just the mean of all actions
    return next_pos, reward, target_state_reached


  def get_next_state_position( self, x, y, direction ):
    ''' given the current state position and direction calculate the postion of the next state '''
    next_pos = []
    if direction==Direction.North or direction == 'N': next_pos = [x,y-1]
    elif direction==Direction.South or direction == 'S': next_pos = [x,y+1]
    elif direction==Direction.East or direction == 'E': next_pos = [x+1,y]
    elif direction==Direction.West or direction == 'W': next_pos = [x-1,y]
    else: next_pos = [x,y] # stay in the current state
    return next_pos


  def get_reward( self, x, y, direction = None ):
    ''' get the reward for moving to the cell at (x,y) or, if a direction is specified,
        the reward for moving from the cell at (x,y) to the cell in the specified direction
    '''
    if direction:

      # if the direction is given as an enum convert to char
      if type(direction) != str:
        direction = Direction.get_direction_char(direction)

      # calculate the postion of the next state
      nx,ny = self.get_next_state_position( x, y, direction )
      return self.grid_base.get_reward(nx,ny),[nx,ny]

    # get the reward for taking this action
    return self.grid_base.get_reward(x,y)


  '''
      Graphical Functions
  '''

  def draw( self ):
    ''' render the grid canvases '''
    return self.draw_grid.canvases

  def clear( self, all_info=False ):
    ''' clear anything currently in the info panels '''
    self.draw_grid.clear(all_info)

  def show_info( self, info: dict ):
    ''' add the supplied information to the grid '''
    self.draw_info.draw( info )

  def save( self, filename ):
    ''' render the grid canvases '''
    canvases = self.get_canvases()
    # save and restore to complete any canvas drawing
    canvases[3].save()
    canvases[3].restore()
    return canvases.to_file(filename)

  def get_canvases(self):
    ''' get the grid levels multi-canvas '''
    return self.draw_grid.canvases

  def get_canvas_dimensions( self ):
    ''' get the total size of the grid canvas in pixels '''
    return [self.draw_grid.total_width, self.draw_grid.total_height]



import numpy as np

import gymnasium
from gymnasium.spaces import Discrete

from .baby_robot_env_v3 import BabyRobotEnv_v3
from .lib.direction import Direction
from .lib.actions import Actions


class Dynamic(gymnasium.Space):

  def __init__(self, action_list = []):
      ' set the list of initially available actions '
      self.set_actions(action_list)

  def sample(self):
      ' select a random action from the set of available actions '
      return np.random.choice(self.available_actions)

  def set_actions(self,actions):
      self.available_actions = actions
      self.n = len(actions)

  def get_available_actions(self):
      return [str(action) for action in self.available_actions]



class BabyRobotEnv_v4( BabyRobotEnv_v3 ):

  def __init__(self, **kwargs):
      super().__init__(**kwargs)

      # initially no actions are available
      self.dynamic_action_space = Dynamic()

      # by default use a dynamic action space
      if kwargs.get('action_space','dynamic') == 'dynamic':
        self.action_space = self.dynamic_action_space
      else:
        # use discrete action space
        # - required for Stable Baselines environment checker which cant yet
        #   recognise dynamic spaces
        # - all actions are available in each state
        # - there are 5 possible actions: move N,E,S,W or stay in same state
        self.action_space = Discrete(5)

      # set the initial position and available actions
      self.reset()


  def get_available_actions( self ):
      ''' test which actions are allowed at the specified grid state '''

      # get the available actions from the grid level
      direction_value = self.level.get_directions(self.x,self.y)

      # convert the grid directions into environment actions
      action_list = []
      if direction_value & Direction.North: action_list.append( Actions.North )
      if direction_value & Direction.South: action_list.append( Actions.South )
      if direction_value & Direction.East:  action_list.append( Actions.East )
      if direction_value & Direction.West:  action_list.append( Actions.West )
      return action_list


  def set_available_actions( self ):
      ' set the list of available actions into the action space '
      action_list = self.get_available_actions()
      self.dynamic_action_space.set_actions( action_list )


  def show_available_actions( self ):
      ''' print the set of avaiable actions for current state '''
      available_actions = str(self.dynamic_action_space.get_available_actions()).replace("'","")
      print(f"({self.x},{self.y}) {available_actions:29}",end="")


  def take_action(self, action):
      ''' apply the supplied action '''

      # call the parent class to take the action and update the position
      super().take_action( action )

      # set the available actions for the new state
      self.set_available_actions()


  def reset(self, seed=None, return_info=False, options=None):
      # reset Baby Robot's position in the grid
      observation,info = super().reset(seed=seed)
      self.set_available_actions()
      return observation,info



# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.

from ..envs.lib.grid_level import GridLevel
from ipywidgets import Layout
from ipywidgets import Play, IntProgress, HBox, VBox, link
import imageio
import os
import gymnasium


class Utils():

  def setup_play_level( level:GridLevel, on_update, interval=1000, min=0, max=8 ):
    ''' setup all the main components required to animate a grid level '''    
    play = Play(interval=interval, min=min, max=max, step=1)
    progress = IntProgress(min=min, max=max)

    link((play, 'value'), (progress, 'value'))
    play.observe(on_update, 'value')

    canvas_dimensions = level.get_canvas_dimensions()
    layout = Layout(width=f'{canvas_dimensions[0]}px')
    return play, progress, layout


  def create_movie( movie_name, image_folder, max_steps, duration = 1.0 ):
    ''' create a gif movie from the images in the specified directory
        where:
          duration = time between each frame
    '''
    with imageio.get_writer(movie_name, mode='I', duration=duration) as writer:
      for index in range(0,max_steps):
        file = f"{image_folder}/step_{index}.png"
        if os.path.exists(file):
          image = imageio.imread(file)
          _ = writer.append_data(image)


  def clear_directory( image_folder ):
    ''' clear any existing files from the directory '''
    filelist = [ f for f in os.listdir(image_folder) if f.endswith(".png") ]
    for f in filelist:
        os.remove(os.path.join(image_folder, f))


  def create_image_directory( image_folder ):
    ''' create the specified folder to store each of the images that
        form an animated gif
        - if the folder already exists then clear any current files
    '''
    if not os.path.exists(image_folder):
        os.makedirs(image_folder)

    # clear any existing files from the directory
    Utils.clear_directory( image_folder )


def make( id: str, render_mode='human', **setup: dict ):
  ''' custom make function for BabyRobot (used instead of 'gym.make')

    * In the latest version of Gym it forces the 'reset' function to be called
      before 'render'. This gives the error:
      "Cannot call `env.render()` before calling `env.reset()`")

      Since we want to just create and then draw the environment
      we want to turn this off (using "env._disable_render_order_enforcing=True")

    * The render mode must be supplied to the make operation, otherwise you get
      the warning:

      "You are calling render method, but you didn't specified the argument render_mode
      at environment initialization."

    * The 'step' function now, instead of returning a single 'done' boolean to indicate
      the end of the episode, returns 2 booleans, 'terminated' and 'truncated'.
      The 'apply_api_compatibility=False' parameter is required to stop a warning appearing:
      "Initializing wrapper in old step API"

    * id: The string used to create the environment with `gym.make`
  '''

  # by default run Baby Robot in Jupyter Notebook graphical mode
  setup['render_mode'] = render_mode

  # if 'max_episode_steps' is set the '_disable_render_order_enforcing' value is not
  # respected - therefore convert to a custom parameter and remove
  if 'max_episode_steps' in setup:
    setup['max_steps'] = setup['max_episode_steps']
    del setup['max_episode_steps']

  env = gymnasium.make(id, **setup)
  env._disable_render_order_enforcing=True
  return env




from .baby_robot_env_v4 import BabyRobotEnv_v4


class BabyRobotEnv_v5( BabyRobotEnv_v4 ):
  
  def __init__(self, **kwargs):
      super().__init__(**kwargs)   
  
  def render(self, mode='human', info=None ):                 
      ''' render as an HTML5 canvas '''
            
      # move baby robot to the current position
      self.robot.move(self.x,self.y) 

      # write the info to the grid side-panel      
      self.level.show_info(info) 

      return self.level.draw()  

  def show_info(self,info):
      ''' display the supplied information on the grid level '''
      self.level.show_info( info )

  def clear_info(self,all_info=False):
      ''' clear any current information of the grid level '''
      self.level.clear(all_info)       


import gymnasium


class BabyRobotEnv_v0(gymnasium.Env):
    
    def __init__(self):
        super().__init__()
        pass

    def step(self, action):        
        state = 1    
        reward = -1            
        terminated = True
        truncated = False
        info = {}
        return state, reward, terminated, truncated, info

    def reset(self, seed=None, options=None):
        super().reset(seed=seed)
        state = 0
        info = {}
        return state,info
  
    def render(self):
        pass




import os
from enum import IntEnum
from .maze import Maze
import numpy as np
from typing import Union


class Puddle(IntEnum):
    Dry, Small, Large = range(3)


class GridBase():

  maze = None                # instance of maze if defined
  puddles = None             # set of tiles where puddles exist
  base_areas = []            # any areas that exist on the base
  grid_areas = []            # and areas that exist on the grid

  debug_maze = False         # write the maze to a svg file

  grid_rewards = []          # the rewards calculated for the whole grid


  def __init__( self, working_directory: str = ".", **kwargs: dict ):

    self.working_directory = working_directory

    # test if a special drawing mode should be used
    # - currently required to run on Google Colab
    # - try to automatically detect if running in Colab using environ
    self.drawmode = kwargs.get('drawmode', 'colab' if 'COLAB_GPU' in os.environ else "" )

    self.width = kwargs.get('width',3)
    self.height = kwargs.get('height',3)

    # the start and end positions in the grid
    # - by default these are the top-left and bottom-right respectively
    self.start = kwargs.get('start',[0,0])
    self.end = kwargs.get('end',[self.width-1,self.height-1])

    # setup any puddles
    self.puddles = kwargs.get('puddles',None)

    # setup up any properties defined for the puddles
    puddle_props = kwargs.get('puddle_props',{})
    self.large_puddle_reward = puddle_props.get('large_reward',-4)
    self.small_puddle_reward = puddle_props.get('small_reward',-2)
    self.large_puddle_probability = puddle_props.get('large_prob',0.4)
    self.small_puddle_probability = puddle_props.get('small_prob',0.6)

    # setup any base-level areas
    self.base_areas = kwargs.get('base_areas',[])

    # setup any grid-level areas
    self.grid_areas = kwargs.get('grid_areas',[])

    # setup any maze and walls
    self.add_maze = kwargs.get('add_maze',False)
    self.maze_seed = kwargs.get('maze_seed',0)
    self.make_maze()
    self.toggle_walls( kwargs.get('walls',[]) )

    # calculate the rewards for each cell in the grid
    self.grid_rewards = self.get_reward()


  '''
      Maze and Walls
  '''

  def make_maze(self):
    if self.add_maze:
      if self.maze is None:
        self.maze = Maze(self.width, self.height, self.start[0], self.start[1], seed = self.maze_seed)
        self.maze.make_maze()
      if self.debug_maze:
        self.maze.write_svg(os.path.join(self.working_directory, "maze.svg"))


  def toggle_walls(self, walls):
    ''' add or remove the specified walls from the grid '''

    # if a maze isnt already defined begin with a maze with no walls
    if self.maze is None:
      self.maze = Maze(self.width, self.height, self.start[0], self.start[1], no_walls = True)
      self.add_maze = True # we now have a maze to add to the canvas

    for (loc), direction in walls:
      x = loc[0]
      y = loc[1]
      num_cells = loc[2] if len(loc) == 3 else 1

      for n in range(num_cells):

        if x >= self.width or y >= self.height:
          break

        current_cell = self.maze.cell_at(x,y)
        if   direction == 'E': next_cell = self.maze.cell_at(x+1,y)
        elif direction == 'W': next_cell = self.maze.cell_at(x-1,y)
        elif direction == 'N': next_cell = self.maze.cell_at(x,y-1)
        elif direction == 'S': next_cell = self.maze.cell_at(x,y+1)

        # add a new wall if none already otherwise remove
        current_cell.toggle_wall(next_cell, direction)

        # move to the next cell for wall repeated across multiple cells
        if direction == 'E' or direction == 'W': y += 1
        else: x += 1


  '''
      Puddles
  '''

  def get_puddle_size( self, x, y ):
    ''' get the size of the puddle at the supplied location '''
    if self.puddles is not None:
      if isinstance(self.puddles[0],list):
        return Puddle(self.puddles[y][x])
      else:
        for (px,py),puddle_size in self.puddles:
          if x==px and y==py:
            return Puddle(puddle_size)

    return Puddle.Dry


  def get_transition_probability( self, x, y ):
    ''' get the probability of moving to the target state when starting in the state at (x,y) '''
    puddle_size = self.get_puddle_size( x, y )

    if puddle_size == Puddle.Large: return self.large_puddle_probability
    if puddle_size == Puddle.Small: return self.small_puddle_probability

    # if no puddle then guaranteed to reach target
    return 1.


  def get_reward( self, x: int = None, y: int = None ) -> Union[int,np.ndarray]:
    ''' return the reward for the specified grid cell '''
    if (x is None) or (y is None):
      return self.get_reward_array()
    else:
      return self.get_reward_value(x,y)


  def get_reward_array(self) -> np.ndarray:
    ''' return a numpy array containing the reward value for all grid cells '''

    if len(self.grid_rewards) == 0:
      height = self.height
      width = self.width
      reward_arr = np.zeros((height,width)).astype(int)
      for y in range(height):
        for x in range(width):
          reward_arr[y][x] = self.get_reward_value(x,y)
      return reward_arr

    # grid rewards already calculated
    return self.grid_rewards


  def test_for_base_area( self, x, y ):
    ''' test if the specified cell is in a base area '''
    for area in self.base_areas:
      # test if only the area defn has been supplied
      if type(area[0]).__name__ == 'int':
        ax,ay,aw,ah = self.get_area_defn(area)
      else:
        ax,ay,aw,ah = self.get_area_defn(area[0])
      
      if self.in_area(x,y,ax,ay,aw,ah):
        return True
        
    return False


  def get_reward_value( self, x, y ):
    ''' return the reward obtained for moving to the specified state

        The amount of reward is a function of the puddle size:
        - no puddle = -1
        - small puddle = -2
        - large puddle = -4

        Actions taken in the terminal state have a reward of zero (although once the terminal
        state is reached the episode terminates, so no actions will occur)

        This represents the amount of time required to move through the puddle (and therefore
        the amount of energy used by BabyRobot)
    '''

    if len(self.grid_rewards) > 0:
      # use the pre-calculated rewards
      # - note this holds the rewards as row,col so must swap x and y
      return self.grid_rewards[y,x]

    puddle_size = self.get_puddle_size( x, y )
    if   puddle_size == Puddle.Large: return self.large_puddle_reward
    elif puddle_size == Puddle.Small: return self.small_puddle_reward

    # no rewards exist off the grid
    if self.test_for_base_area(x,y):
      return 0

    # by definition the terminal state has a zero reward
    if (x == self.end[0]) and (y == self.end[1]):
      return 0

    # if any grid areas exist these can set different rewards
    # - the most recently defined area is the one whose reward will be taken
    # for a cell
    cell_reward = -1
    for area in self.grid_areas:
      if len(area) > 2:
        try:
          ax,ay,aw,ah = self.get_area_defn(area[0])
          if self.in_area( x,y,ax,ay,aw,ah ):
            cell_reward = area[2]
        except:
          pass

    return cell_reward

  '''
     Areas
  '''

  def get_area_defn( self, area):
    ''' extract the area properties from the supplied area tuple '''
    x,y,*args = area
    wd,ht = args if args else (1,1) # default to unit-square if no values specified
    return x,y,wd,ht


  def in_area(self,x,y,ax,ay,aw,ah):
    if (x >= ax and x < (ax + aw)) and ((y >= ay and y < (ay + ah))):
      return True
    return False



from ipycanvas import Canvas, hold_canvas
from ipywidgets import Image
from time import sleep

import random
import os

from .robot_position import RobotPosition
from .draw_grid import Level


class RobotDraw( RobotPosition ):

  sprite_count = 0
  canvas_sprites = []

  x_offset = 0
  y_offset = 0


  def __init__( self, level, **kwargs ):
      super().__init__( level, **kwargs )

      # the number of draw steps since the last reset
      self.move_step = 0

      # set to use the robot level of the multicanvas
      self.canvas = self.grid.canvases[Level.Robot]

      # get any robot specific parameters
      robot_params = kwargs.get('robot',{})

      # the number of steps before a sprite change
      self.sprite_change = robot_params.get('sprite_change',2)

      # the offset of the grid display
      offset = kwargs.get('offset',[0,0])
      self.x_offset = offset[0]
      self.y_offset = offset[1]

      # test that baby robot should be shown
      self.show_robot = robot_params.get('show',True)

      # get robot speed parameters
      self.sleep = robot_params.get('sleep',0.07)
      self.canvas_sleep = robot_params.get('canvas_sleep',40)

      # the initial robot sprite (4 = center view)
      self.sprite_index = robot_params.get('initial_sprite',4)
      self.load_sprites()


  def get_number_of_sprites(self):
      ''' return the number of sprites on the sprite sheet '''
      num_sprites = len(self.canvas_sprites)
      return num_sprites


  def load_single_sprint(self):
      ''' load the defined sprite index - default is center robot '''
      image_path = os.path.join(self.level.working_directory,f'images/baby_robot_{self.sprite_index}.png')
      self.sprite = Image.from_file(image_path)


  def add_sprite(self,index):
      ''' add an image sprite with the specified index from an individual image
          to the set of sprites to use for robot drawing
      '''
      # currently using individual sprite images rather than sprite sheet to fix Colab
      image_path = os.path.join(self.level.working_directory, f'images/baby_robot_{index}.png')
      sprites = Image.from_file(image_path)
      # put all sprites onto their own canvas
      canvas = Canvas(width=self.robot_size, height=self.robot_size)
      canvas.draw_image(sprites, 0, 0)
      self.canvas_sprites.append(canvas)


  def load_sprites(self):
      ''' load the sprites used to draw baby robot
          - on Colab it seems to be limited to a single image and not on a canvas
      '''
      self.canvas_sprites = []
      if self.level.drawmode == 'colab':
        self.load_single_sprint()
      else:
        # load all sprites
        for row in range(5):
          for col in range(2):
            index = (row*2) + col
            self.add_sprite(index)
        self.add_sprite(index+1)


  def update_sprite(self):
      ' randomly change to the next sprite '

      if self.get_number_of_sprites() > 1:
        self.sprite_count += 1
        if self.sprite_count > self.sprite_change:
            self.sprite_count = 0
            self.sprite_index = self.sprite_index + random.randint(-1,+1)
            if self.sprite_index < 0:
                self.sprite_index = 0
            if self.sprite_index >= self.get_number_of_sprites():
                self.sprite_index = (self.get_number_of_sprites()-1)


  def draw_sprite(self,index):
      ' remove the last sprite and add the new one at the current position '

      x = self.x + self.x_offset
      y = self.y + self.y_offset

      if self.level.drawmode == 'colab':
        self.canvas.clear_rect(x-10, y-10, self.robot_size+10, self.robot_size+10)
        self.canvas.draw_image(self.sprite, x, y )

      elif self.sprite_index < self.get_number_of_sprites():
        with hold_canvas(self.canvas):
          self.canvas.clear_rect(x, y, self.robot_size)
          self.canvas.draw_image(self.canvas_sprites[index], x, y )


  def draw(self):
      ' add the current sprite at the current position '
      if self.show_robot:
        self.draw_sprite(self.sprite_index)
        self.update_sprite()


  def move_direction(self,direction):
      ' move from one square to the next in the specified direction '

      # test if a move is actually specified
      if self.test_for_valid_move(direction):
          move_method_name = f"move_{direction.name}"

          if self.level.drawmode == 'colab':
            self.move_direction_colab( move_method_name )
          else:
            for _ in range(self.robot_size//self.step):
              getattr(self,move_method_name)()
              self.draw()
              sleep(self.sleep) # pause between each move step
              self.move_step += 1

          self.move_count += 1
      else:
          # not moving just pausing = action.Stay
          sleep(self.sleep * (self.robot_size//self.step))


  def move_direction_colab( self, move_method_name ):
      ''' move from one square to the next in the specified direction
          - special drawing method to overcome problems in Colab
      '''
      with hold_canvas(self.canvas):
        for _ in range(self.robot_size//self.step):
            getattr(self,move_method_name)()
            self.canvas.clear_rect(self.x-10, self.y-10, self.robot_size+10, self.robot_size+10)
            self.canvas.draw_image(self.sprite, self.x, self.y )
            self.canvas.sleep(self.canvas_sleep)
            sleep(self.sleep) # pause between each move step
            self.move_step += 1


  def reset(self):
      ' clear the robot layer '
      self.move_step = 0
      self.canvas.clear()
      self.set_cell_position(self.initial_position)
      self.draw()


  def partial_move(self,direction,sprite_index=None):
      ' move from one square to the next in the specified direction '

      if direction is not None:
        if self.test_for_valid_move(direction):
          move_method_name = f"move_{direction.name}"
          getattr(self,move_method_name)()

          if sprite_index is not None:
            self.sprite_index = sprite_index
          self.draw()
          self.move_count += 1

          # a move took place
          return True

      # no move occurred
      return False


import gymnasium

from .lib.grid_level import GridLevel
from .lib.robot import Robot
from .lib.dynamic_space import Dynamic
from .lib.direction import Direction

class BabyRobotBase(gymnasium.Env):
    ''' Baby Robot Gym Environment Base Class '''

    def __init__(self, **kwargs):
        super().__init__()


# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.


import random
import numpy as np
from gymnasium.spaces import Discrete, MultiDiscrete
from .baby_robot_interface import BabyRobotInterface


class BabyRobot_v0( BabyRobotInterface ):
    ''' Baby Robot Gym Environment '''

    metadata = {'render_modes': ['human'], 'render_fps': 4}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # test if the 'step' function should use the new, 2 boolean, return format
        self.apply_api_compatibility = kwargs.get('apply_api_compatibility',False)

        self.max_episode_steps = kwargs.get('max_steps', None)

        # by default use a dynamic action space
        if kwargs.get('action_space','dynamic') == 'dynamic':
          self.action_space = self.dynamic_action_space
        else:
          # use discrete action space
          # - required for Stable Baselines environment checker which can't yet
          #   recognise dynamic spaces
          # - all actions are available in each state
          # - there are 5 possible actions: move N,E,S,W or stay in same state
          self.action_space = Discrete(5)

        # the observation will be the coordinates of Baby Robot
        self.observation_space = MultiDiscrete([self.width, self.height])


    #
    # Gym Interface Methods
    #

    def step(self, action):
        ''' take the action and update the position

            - under the latest version of Gym a single 'done' value is no longer returned, instead 2 boolean values are used:
            (from the Gym documentation)

            terminated (bool): whether a `terminal state` (as defined under the MDP of the task) is reached.
                In this case further step() calls could return undefined results.
            truncated (bool): whether a truncation condition outside the scope of the MDP is satisfied.
                Typically a timelimit, but could also be used to indicate agent physically going out of bounds.
                Can be used to end the episode prematurely before a `terminal state` is reached.

        '''
        reward, target_reached = self.take_action(action)
        obs = np.array([self.x,self.y])

        # increment the number of steps taken since the last reset
        # - if this is greater than the maximum allowed for the episode set the 'truncated' flag
        self.steps += 1
        truncated = False if self.max_episode_steps is None else (self.steps > self.max_episode_steps)
        
        # set the 'terminated' flag if we've reached the exit or the episode is truncated
        # (previously this was called 'done')
        terminated = (self.x == self.end[0]) and (self.y == self.end[1])        
        if truncated: terminated = True

        info = {'target_reached':target_reached}

        if self.apply_api_compatibility:
          # old style return format - uses a single boolean to indicate episode termination
          return obs, reward, terminated, info
        else:
          # new style return format - uses 2 booleans
          return obs, reward, terminated, truncated, info


    def render(self, mode='human', info=None ):
        ''' render as an HTML5 canvas '''
        # move baby robot to the current position
        self.robot.move(self.x,self.y)

        # write the info to the grid side-panel
        self.level.show_info(info)
        return self.level.draw()


    def reset(self, seed=None, return_info=False, options=None):
        ''' reset Baby Robot's position in the grid '''
        super().reset(seed=seed)
        if seed is not None:
           random.seed(seed)

        # the number of steps taken since the last reset
        self.steps = 0

        self.robot.set_cell_position(self.initial_pos)
        self.robot.reset()
        self.x = self.initial_pos[0]
        self.y = self.initial_pos[1]
        self.set_available_actions()
        info = {}
        return np.array([self.x,self.y]),info





from .grid_level import GridLevel
from .draw_grid import DrawGrid
from .draw_info import DrawInfo


class GraphicalGridLevel( GridLevel ):

  def __init__( self, **kwargs: dict ):
    super().__init__( **kwargs )
    self.draw_grid = DrawGrid( self.grid_base, **kwargs )
    self.draw_info = DrawInfo( self.draw_grid, self.grid_info, **kwargs )    


# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.

import os
os.environ["JUPYTER_PLATFORM_DIRS"] = "1"

#
# Release Gym Environments
#

from .baby_robot_interface import BabyRobotInterface
from .baby_robot_v0 import BabyRobot_v0



#
# Example Gym Environments
#

from .baby_robot_env_v0 import BabyRobotEnv_v0
from .baby_robot_env_v1 import BabyRobotEnv_v1
from .baby_robot_env_v2 import BabyRobotEnv_v2
from .baby_robot_env_v3 import BabyRobotEnv_v3
from .baby_robot_env_v4 import BabyRobotEnv_v4
from .baby_robot_env_v5 import BabyRobotEnv_v5
from .baby_robot_env_v6 import BabyRobotEnv_v6
from .baby_robot_env_v7 import BabyRobotEnv_v7




import numpy as np
from typing import Union


from ipycanvas import hold_canvas
from .arrows import Arrows
from .direction import Direction

from .grid_base import GridBase


''' class to write information to a grid level '''
class GridInfo():

  def __init__(self, gridbase: GridBase, **kwargs: dict):
    self.grid = gridbase

  def get_cell_directions(self, x: int, y: int, direction: Direction = None) -> list:
    ''' return the list of available directions for the specified position in the grid
        returns a list in the form: {'N':True,'E':False,'S':False,'W':True}
    '''

    grid = self.grid

    # no actions exist for the terminal state
    if (x == grid.end[0]) and (y == grid.end[1]):
      return {}

    # no actions exist off the grid
    for area in grid.base_areas:
      # test if only the area defn has been supplied
      if type(area[0]).__name__ == 'int':
        ax,ay,aw,ah = grid.get_area_defn(area)
      else:
        ax,ay,aw,ah = grid.get_area_defn(area[0])
      # if (x >= ax and x < (ax + aw)) and ((y >= ay and y < (ay + ah))):
      #   return {}
      if grid.in_area(x,y,ax,ay,aw,ah):
        return {}

    # test if the level contains a maze
    if grid.maze is not None:
      cell = grid.maze.cell_at( x, y )

      # if a wall is present then that direction is not possible as an action
      actions = {k: not v for k, v in cell.walls.items()}
    else:
      # initially start with all actions being possible
      actions = {'N':True,'E':True,'S':True,'W':True}

      # remove actions that would move off the edges of the grid
      if x == 0: del actions['W']
      if x == grid.width-1: del actions['E']
      if y == 0: del actions['N']
      if y == grid.height-1: del actions['S']

    # remove actions that would move to an off-grid cell
    for area in grid.base_areas:
      # test if only the area defn has been supplied
      if type(area[0]).__name__ == 'int':
        ax,ay,aw,ah = grid.get_area_defn(area)
      else:
        ax,ay,aw,ah = grid.get_area_defn(area[0])

      if grid.in_area(x+1,y,ax,ay,aw,ah): del actions['E']
      if grid.in_area(x-1,y,ax,ay,aw,ah): del actions['W']
      if grid.in_area(x,y-1,ax,ay,aw,ah): del actions['N']
      if grid.in_area(x,y+1,ax,ay,aw,ah): del actions['S']


    if direction is not None:
      # set any allowed actions to false if they dont match the supplied action
      dir_value = direction
      for dir,v in actions.items():
        if v == True:
          if (dir == 'N') and not (dir_value & Direction.North): actions['N'] = False
          if (dir == 'S') and not (dir_value & Direction.South): actions['S'] = False
          if (dir == 'E') and not (dir_value & Direction.East):  actions['E'] = False
          if (dir == 'W') and not (dir_value & Direction.West):  actions['W'] = False

    return actions


  def get_directions( self, x: int = None, y: int = None ) -> Union[Direction,np.ndarray]:
    ''' return the bitfield value representing the possible directions for the specified grid cell '''
    if (x is None) or (y is None):
      return self.get_direction_array()
    else:
      return self.get_direction_value(x,y)


  def get_direction_value( self, x: int, y: int ) -> Direction:
    ''' return the bitfield value representing the possible directions for the specified grid cell '''
    return Direction.get_value( self.get_cell_directions( x, y ) )


  def get_direction_array(self) -> np.ndarray:
    ''' return a numpy array containing the direction value for all grid cells '''
    height = self.grid.height
    width = self.grid.width
    direction_arr = np.zeros((height,width)).astype(int)
    for y in range(height):
      for x in range(width):
        direction_arr[y][x] = self.get_direction_value(x,y)
    return direction_arr


# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.


from gymnasium.envs.registration import register
from .lib.utils import make
from ._version import __version__


#
# Release Gym Environments
#


register(
    id='BabyRobot-v0',
    entry_point='babyrobot.envs:BabyRobot_v0',
)


#
# Example Gym Environments
#

register(
    id='BabyRobotEnv-v0',
    entry_point='babyrobot.envs:BabyRobotEnv_v0',
)

register(
    id='BabyRobotEnv-v1',
    entry_point='babyrobot.envs:BabyRobotEnv_v1',
)

register(
    id='BabyRobotEnv-v2',
    entry_point='babyrobot.envs:BabyRobotEnv_v2',
)

register(
    id='BabyRobotEnv-v3',
    entry_point='babyrobot.envs:BabyRobotEnv_v3',
)

register(
    id='BabyRobotEnv-v4',
    entry_point='babyrobot.envs:BabyRobotEnv_v4',
)

register(
    id='BabyRobotEnv-v5',
    entry_point='babyrobot.envs:BabyRobotEnv_v5',
)

register(
    id='BabyRobotEnv-v6',
    entry_point='babyrobot.envs:BabyRobotEnv_v6',
)

register(
    id='BabyRobotEnv-v7',
    entry_point='babyrobot.envs:BabyRobotEnv_v7',
)




import numpy as np
from numpy import inf
from ipycanvas import hold_canvas
from .draw_grid import DrawGrid, Level
from .grid_info import GridInfo
from .arrows import Arrows
from .direction import Direction


class DrawInfo():

  arrow_color = '#00008b'    # color of any information arrows

  text_bg_color = 'rgba(40,40,40,0.7)' # text background shading
  text_fg_color = '#fff'               # text foreground color

  zero_fg_color = text_fg_color        # zero value text background shading
  zero_bg_color = text_bg_color        # zero value text foreground color

  precision = 3   # the precision to use when writing floating point values


  def __init__( self, draw_grid: DrawGrid, grid_info: GridInfo, **kwargs: dict ):

    # get the basic details of the grid
    self.grid = draw_grid.grid

    # store the GridInfo object
    self.grid_info = grid_info

    # get the underlying grid drawing and canvases
    self.draw_grid = draw_grid
    self.canvas = self.draw_grid.canvases[Level.Text]

    # setup the grid properties
    self.set_properties(kwargs.get('grid',None))

    # arrow draw class
    self.arrows = Arrows( draw_grid.cell_pixels, draw_grid.padding,length=24,width=7,height=11)


  '''
      Interface Functions
  '''

  def draw( self, props: dict ):
    ''' add the supplied info to the grid '''

    # test if any information has been supplied
    # - if it has then print this as a string
    if props is not None:

      # test if a floating point precision has been defined
      self.precision = props.get('precision', self.precision)

      # test if any directions have been specified to add to the grid
      directions = props.get('directions',None)
      if directions is not None:
        self.process_direction_arrows(directions)
        self.process_direction_text(directions)

      # process any text to add to the grid
      self.process_text(props)

      # process anything to add to the info panel
      self.process_info(props)

      # test if coordinates should be added to the grid
      if props.get('coords',False):
        self.draw_coordinates()


  '''
      Info Dictionary Processing
  '''

  def set_default_values(self,args):
    ''' replace any missing tuple elements in a text item with a default value '''

    # the default values for the info panel text item
    defaultargs = ((0,0),"",190,20)

    # pad args with None up to the length of the default tuple
    args += (None,)*len(defaultargs)

    # process each element & replace missing elements with default
    args = tuple(map(lambda x, y: y if y is not None else x, defaultargs, args))
    return args


  def process_info(self,info):
      ''' test if any text has been specified to add to the info panel '''

      # default colors to use when writing info
      fg_color = 'black'
      bk_color = 'white'

      text = info.get('side_info',None)
      if text is not None:

        # check that a side panel was setup during grid creation
        if self.draw_grid.side_panel is None:
          raise Exception("\'side_panel\' must be specified during grid creation to allow side panel text.")

        # setup the info panel colors
        if type(self.draw_grid.side_panel) == dict:
          fg_color = self.draw_grid.side_panel.get('text_fg','black')
          bk_color = self.draw_grid.side_panel.get('color','white')

        # put text on side panel
        if self.draw_grid.side_panel:

          # clear existing text
          for item in text:
            (cx,cy),value,width,height = self.set_default_values(item)

            # offset by the grid width to get the start of the side panel
            cx += self.draw_grid.width_pixels
            self.clear_info_panel_text( cx, cy, width, height, bk_color)

          for item in text:
            (cx,cy),value,width,height = self.set_default_values(item)

            # offset by the grid width to get the start of the side panel
            cx += self.draw_grid.width_pixels
            self.info_panel_text(cx,cy,value,width,height,fg_color=fg_color,bk_color=bk_color)


      text = info.get('bottom_info',None)
      if text is not None:

        # check that a side panel was setup during grid creation
        if self.draw_grid.bottom_panel is None:
          raise Exception("\'bottom_panel\' must be specified during grid creation to allow bottom panel text.")

        if type(self.draw_grid.bottom_panel) == dict:
          fg_color = self.draw_grid.bottom_panel.get('text_fg','black')
          bk_color = self.draw_grid.bottom_panel.get('color','white')

        # put text on bottom panel
        if self.draw_grid.bottom_panel:

          # clear existing text
          for item in text:
            (cx,cy),value,width,height = self.set_default_values(item)
            cy += self.draw_grid.height_pixels
            self.clear_info_panel_text( cx, cy, width, height, bk_color)

          for item in text:
            (cx,cy),value,width,height = self.set_default_values(item)

            # offset by the grid height to get the start of the bottom panel
            cy += self.draw_grid.height_pixels
            self.info_panel_text(cx,cy,value,width,height,fg_color=fg_color,bk_color=bk_color)


  def process_text(self,info):
      ''' test if any text has been specified to add to the grid '''

      text = info.get('text',None)
      if text is not None:
        if isinstance(text,np.ndarray):
          self.draw_text_array(text)
        else:
          for (cx,cy),value in text:
            self.draw_cell_text(cx,cy,value)

      values = info.get('values',None)
      if values is not None and isinstance(values,np.ndarray):
        if len(values.shape) == 3:
          if values.shape[2] == 5:
            values = values[:,:,1:] # remove the 'stay' action values
          self.draw_text_array(values)



  def process_direction_arrows(self,directions):
    ''' test if any arrows should be added to the grid '''

    arrows = directions.get('arrows',None)
    if arrows is not None:
      # directions can be specified as a numpy array containing the
      # directions for all cells or as individual cell directions
      if isinstance(arrows,np.ndarray):
        self.draw_direction_arrow_array(arrows)
      else:

        if len(arrows) > 0 and (type(arrows[0][0]) == int):
          # only cell coordinate supplied
          for (cx,cy) in arrows:
            direction = self.grid_info.get_directions(cx,cy)
            self.draw_direction_arrow(cx,cy,direction)

        else:
          # direction supplied with each cell
          for (cx,cy),direction in arrows:
            self.draw_direction_arrow(cx,cy,direction)


  def process_direction_text(self,directions):
    ''' test if the direction text should be added to the grid '''

    text = directions.get('text',None)
    if text is not None:
      # direction text can be specified as a numpy array containing the
      # directions for all cells or as individual cell directions
      if isinstance(text,np.ndarray):
        self.draw_direction_text_array(text)
      else:
        if len(text) > 0 and (type(text[0][0]) == int):
          # only cell coordinate supplied
          for (cx,cy) in text:
            direction = self.grid_info.get_directions(cx,cy)
            self.draw_direction_text(cx,cy,direction)

        else:
          # direction supplied with each cell
          for (cx,cy),direction in text:
            self.draw_direction_text(cx,cy,direction)


  '''
      Setup Functions
  '''

  def set_properties( self, grid_props: dict ):
    ''' setup the draw info properties '''

    if grid_props is not None:
      colors = grid_props.get('colors',None)
      if colors is not None:
        self.arrow_color = colors.get('arrows', self.arrow_color)
        self.text_fg_color = colors.get('text_fg', self.text_fg_color)
        self.text_bg_color = colors.get('text_bg', self.text_bg_color)

        # colors for zero values - by default the same as standard text
        self.zero_fg_color = colors.get('zero_fg', self.text_fg_color)
        self.zero_bg_color = colors.get('zero_bg', self.text_bg_color)


  '''
      Direction Arrow Functions
  '''

  def draw_direction_arrow( self, x, y, directions ):
    ''' draw an arrow in each direction from the supplied list '''

    canvas = self.draw_grid.canvases[Level.Overlay]
    color = self.arrow_color
    padding = self.draw_grid.padding
    cell_pixels = self.draw_grid.cell_pixels
    px,py = self.draw_grid.grid_to_pixels( [x,y], padding, padding )

    canvas.clear_rect(px,py,cell_pixels,cell_pixels)
    self.arrows.draw(canvas,px,py,directions,color)


  def draw_direction_arrow_array(self, directions: np.array):
    ''' draw arrows in each direction in the supplied numpy array '''
    canvas = self.draw_grid.canvases[Level.Overlay]
    for y in range(directions.shape[0]):
      for x in range(directions.shape[1]):
        self.draw_direction_arrow( x, y, directions[y,x])


  '''
      Text Directions
  '''

  def draw_direction_text( self, x, y, direction ):
    ''' convert the supplied direction bitfield value to a string, with a character for
        each possible direction, and draw this on the specified cell
    '''
    self.draw_cell_text( x, y, Direction.get_string(direction) )


  def draw_direction_text_array(self,directions):
    ''' add a text string to each cell showing the directions '''

    # self.canvas.save()
    # with hold_canvas(self.canvas):
    for y in range(directions.shape[0]):
      for x in range(directions.shape[1]):
        if x != self.grid.end[0] or y != self.grid.end[1]:
          self.draw_direction_text( x, y, directions[y,x])
    # self.canvas.restore()

  '''
      Coordinates
  '''

  def draw_coordinates(self):
    ''' add the coordinates to each cell '''
    with hold_canvas(self.canvas):
      for y in range(self.draw_grid.grid.height):
        for x in range(self.draw_grid.grid.width):
          self.draw_cell_text( x, y, f"({x},{y})")



  '''
      Text Functions
  '''

  def draw_text_array(self,text):
    ''' draw the supplied array of text items to the grid '''
    with hold_canvas(self.canvas):
      for y in range(text.shape[0]):
        for x in range(text.shape[1]):
          if len(text.shape) == 2:
            self.draw_cell_text( x, y, text[y,x])
          else:
            for z in range(text.shape[2]):
                self.draw_cell_text( x, y, text[y,x,z], pos=z)


  def info_panel_text( self, x, y, text,width,height,
                       fg_color='#000',
                       bk_color='#fff',
                       font='bold 14px sans-serif',
                       text_align='left',
                       text_baseline='top'):
    ''' add information text in the side panel '''
    canvas = self.canvas
    canvas.save()
    with hold_canvas(canvas):
      canvas.fill_style = fg_color
      canvas.text_align = text_align
      canvas.text_baseline = text_baseline
      canvas.font = font
      canvas.fill_text(text, x, y)
    canvas.restore()


  def clear_info_panel_text( self, x, y, width, height, bk_color='#fff'):
    ''' clear the side panel at the specified location '''
    canvas = self.canvas
    with hold_canvas(canvas):
      canvas.fill_style = bk_color
      canvas.fill_rect(x,y-5,width,height)


  def draw_cell_text( self, x, y, value, color = None, back_color = None, pos = None ):
    ''' display the given value in the specified cell '''

    # dont draw anything if no text is supplied
    num_value = False
    if type(value).__name__.startswith('str'):
      if len(value) == 0:
        return
    elif isinstance(x, (int, float, complex)) and not isinstance(x, bool):
      num_value = True
      if np.isnan(value):
        # don't show anything for NaN number values
        return

    # dont display text on base areas
    if self.grid.test_for_base_area(x,y):
      return

    # limit floating point values to the default precision
    if isinstance(value, float):
      value = round(value,self.precision)
      if self.precision == 0:
        # convert to int if set to have no decimal places
        value = value.astype(int)

    canvas = self.canvas
    padding = self.draw_grid.padding

    if color is None: 
      if value==0: color = self.zero_fg_color
      else:        color = self.text_fg_color

    if back_color is None:
      if value==0: back_color = self.zero_bg_color
      else:        back_color = self.text_bg_color

    gx,gy = self.draw_grid.grid_to_pixels( [x,y], padding, padding )
    cx,cy = self.draw_grid.get_center(gx,gy) # calculate the center of this cell

    # dimensions of background text shading
    bk_height = 20
    bk_width = 36

    # use a smaller area when more than one text item is being written to the cell
    if (pos is not None):
      bk_height = 16
      bk_width = 26

    # the width is set for 4 characters - expand if more than this
    if len(str(value)) > 4:
      bk_width += (len(str(value))-4) * 6

    # don't let the background extend outside the cell
    if bk_width > (self.draw_grid.cell_pixels - 4):
      bk_width = (self.draw_grid.cell_pixels - 4)

    # center the text
    x_off = (bk_width//2)
    y_off = (bk_height//2)

    if (pos is not None):
      if pos == 0:
        y_off += (self.draw_grid.cell_pixels // 4) + 4
      elif pos == 1:
        x_off -= (self.draw_grid.cell_pixels // 4)
      elif pos == 2:
        y_off -= (self.draw_grid.cell_pixels // 4) + 4
      elif pos == 3:
        x_off += (self.draw_grid.cell_pixels // 4)

    # use a smaller font size for high precision, since more digits to fit
    font_size = 14
    txt_offy = 5
    txt_offx = 0
    if (num_value and self.precision > 1) or \
       (not num_value and len(str(value)) >= 3):
          font_size = 12
          txt_offy = 4
    elif (pos is not None):
        font_size = 11
        if pos == 0:
          txt_offy = -(self.draw_grid.cell_pixels // 4)
        elif pos == 1:
          txt_offx = (self.draw_grid.cell_pixels // 4)
          txt_offy = 4
        elif pos == 2:
          txt_offy = (self.draw_grid.cell_pixels // 4) + 8
        elif pos == 3:
          txt_offx = -(self.draw_grid.cell_pixels // 4)
          txt_offy = 4
    font_str = f"bold {font_size}px sans-serif"

    canvas.save()

    with hold_canvas(canvas):
      canvas.clear_rect(cx-x_off,cy-y_off,bk_width,bk_height)
      if back_color is not None:
        canvas.fill_style = back_color
        canvas.fill_rect(cx-x_off,cy-y_off,bk_width,bk_height)
    
    with hold_canvas(canvas):
      canvas.fill_style = color
      canvas.text_align = 'center'
      canvas.font = font_str
      canvas.fill_text(f"{value}", cx+txt_offx, cy+txt_offy)

    canvas.restore()


# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.

from .policy import Policy
from .deterministic_policy import DeterministicPolicy
from .policy_evaluation import PolicyEvaluation
from .value_iteration import ValueIteration
from .monte_carlo import MonteCarlo, MonteCarloStateValues, MonteCarloActionValues, DeltaType, MonteCarloGPI
from .utils import Utils
from .animation import Animate



# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.

import numpy as np
from ..envs import BabyRobotInterface
from ..envs.lib.direction import Direction
from . import Policy


''' evaluate a policy '''
class PolicyEvaluation():

  iterations = 0
  policy = None
  discount_factor = 1.0  
  threshold = 1e-3

  def __init__(self, env: BabyRobotInterface, policy: Policy, discount_factor=1.0):
    self.env = env

    # check that a policy has been given to evaluate
    assert policy is not None, "A Policy must be supplied to PolicyEvaluation." 

    self.policy = policy
    self.discount_factor = discount_factor
    self.reset()


  def reset(self):
    self.iterations = 0
    self.reset_start_values()
    self.reset_end_values()


  def reset_start_values(self):
    self.start_values = np.zeros((self.env.height,self.env.width))    


  def reset_end_values(self):
    self.end_values = np.zeros((self.env.height,self.env.width))  


  def get_state_value(self,pos):
    ''' get the currently calculated value of the specified position in the grid '''
    x = pos[0]
    y = pos[1]
    if (x < 0 or x >= self.env.width) or (y < 0 or y >= self.env.height): 
      return 0
    return self.start_values[y,x]


  def calculate_action_value(self,x,y,chosen_action,all_actions):

    # get the possible probabilities, next_states and rewards for this action
    action_probabilities = self.env.get_action_probabilities( x, y, chosen_action )

    action_value = 0
    for probability,next_pos,reward in action_probabilities:      
      # convert the x,y position into row,col      
      action_value += probability * (reward + (self.discount_factor * self.get_state_value( next_pos )))   

    return action_value 


  def calculate_state_value(self,x,y):
    ''' calculate the state value when all actions are equally possible '''

    # get the list of all possible actions in this state
    all_actions = self.env.get_available_actions(x,y)

    # when an action is taken it will either end in target state or one of the other possible states
    # - count the number of states other than the target state
    num_actions = len(all_actions)
    if num_actions > 0:
            
      # calculate the total value for all possible actions in this state
      value = 0
      for chosen_action in all_actions:
        action_value = self.calculate_action_value(x,y,chosen_action,all_actions)       

        # add the action value to the total state value        
        value += action_value        

      # for equal probability of taking an action its just the mean of all actions
      return value/num_actions
    
    # no possible actions
    return 0


  def calculate_policy_state_value(self,x,y):
    ''' calculate the state value for a policy '''             

    # get the dictionary of possible actions and their probabilities in this state   
    actions = self.policy.get_action_probabilities(x,y)   
    all_actions = list(actions.keys())
    probabilities = list(actions.values())     

    # when an action is taken it will either end in target state or one of the other possible states
    # - count the number of states other than the target state
    num_actions = len(all_actions)
    if num_actions > 0:                 
            
      # calculate the total value for all possible actions in this state
      value = 0
      for chosen_action,probability in zip(all_actions,probabilities):
        action_value = self.calculate_action_value(x,y,chosen_action,all_actions)       

        # add the action value, multiplied by the probability of taking that action, to the total state value        
        value += (probability * action_value)
      
      return value
    
    # no possible actions
    return 0  


  def standard_sweep(self):
    ''' calculate the state value for all states '''                
    # calculate the value of all states except the exit
    end = self.env.end
    for y in range(self.env.height):
      for x in range(self.env.width):
        if self.env.level.grid_base.test_for_base_area(x,y) == False:
          if (x != end[0]) or (y != end[1]):          
            if self.policy is None:
              # use stochastic policy
              self.end_values[y,x] = self.calculate_state_value(x,y)   
            else:
              # calculate value under deterministic policy
              self.end_values[y,x] = self.calculate_policy_state_value(x,y)  


  def do_iteration(self):        
    self.start_values = self.end_values   # copy the end values into the start values            
    self.reset_end_values()               # reset the end values        
    self.standard_sweep()                 # sweep all states    
    self.iterations += 1                  # increment the iteration count


  def run_to_convergence(self, max_iterations = 100, threshold = 1e-3):
    ''' run until the values stop changing '''
    for n in range(max_iterations):
      self.do_iteration()
      
      # calculate the largest difference in the state values from the start to end of the iteration
      delta = np.max(np.abs(self.end_values - self.start_values))            
      
      # test if the difference is less than the defined convergence threshold
      if delta < threshold:
        break
    
    # return the number of iterations taken to converge
    return n


  def set_policy(self, policy: Policy):
    ''' set the policy to be evaluated '''
    self.policy = policy
    
    # reset the iterations required to run to convergence on the policy
    self.iterations = 0
    

  def set_discount_factor(self, discount_factor):
    ''' set the discount factor to apply to the future rewards '''
    self.discount_factor = discount_factor


  def get_iterations(self):
    ''' return the number of iterations of policy evalutation that have been performed '''
    return self.iterations


import numpy as np

from .baby_robot_env_v2 import BabyRobotEnv_v2

from .lib.grid_level import GridLevel
from .lib.robot_draw import RobotDraw
from .lib.actions import Actions


''' the first graphical environment '''
class BabyRobotEnv_v3( BabyRobotEnv_v2 ):

  def __init__(self, **kwargs):
      super().__init__(**kwargs)

      # graphical creation of the level
      self.level = GridLevel( **kwargs )

      # add baby robot
      self.robot = RobotDraw(self.level,**kwargs)
      self.robot.draw()

  def reset(self, seed=None, return_info=False, options=None):
      super().reset(seed=seed)
      # reset Baby Robot's position in the grid
      self.robot.set_cell_position(self.initial_pos)
      self.robot.reset()
      self.x = self.initial_pos[0]
      self.y = self.initial_pos[1]
      info = {}
      return np.array([self.x,self.y]),info

  def render(self, action=0, reward=0 ):
      ''' render as an HTML5 canvas '''
      print(f"{Actions(action): <5}: ({self.x},{self.y}) reward = {reward}")

      # move baby robot to the current position
      self.robot.move(self.x,self.y)
      return self.level.draw()


# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.

import numpy as np
from tqdm import tqdm
from enum import IntEnum
import babyrobot
from ..envs import BabyRobotInterface
from ..envs.lib import Actions
from ..lib import Policy


class DeltaType(IntEnum):
    Max, Mean = range(2)



class MonteCarlo():

  last_start_pos = None         # the episode start position
  rewards: np.array             # rewards array to be initialised by child class
  values: np.array              # values array to be initialised by child class

  def __init__(self, policy: Policy, exploring_starts=False, every_visit=False, env=None, **env_setup):

    # store the setup used to define the environment
    self.env_setup = env_setup

    self.exploring_starts = exploring_starts
    self.every_visit = every_visit

    # store the policy used to select actions in the environment
    self.policy = policy
    
    if env is None:
      # create an evaluation version of the environment
      self.env = babyrobot.make("BabyRobot-v0", render_mode=None, **env_setup)
    else:
      # use the supplied environment
      self.env = env


  #
  # Graphical Helper Functions
  #

  def get_graphical_environment(self, text_data):
    ''' create a graphical version of the environment - virtual base method '''
    raise NotImplementedError()

  def show_values(self):
    ''' show the calculated state values '''
    env = self.get_graphical_environment(self.returns)
    return env

  def show_visits(self):
    ''' show the visits to each state '''
    env = self.get_graphical_environment(self.visits)
    return env


  #
  # Monte Carlo Method Functions
  #

  def get_next_start_state(self):
    ''' get the environment starting position for the next episode '''

    # by default begin at the defined start position
    x = self.env.start[0]
    y = self.env.start[1]

    if self.exploring_starts:

      if self.last_start_pos is not None:
        # move horizontally to the next state
        x = self.last_start_pos[0] + 1
        y = self.last_start_pos[1]

      while self.env.level.is_grid_state(x,y) == False:
        # move horizontally to the next state
        x += 1

        if x >= self.env.width:
          # move to the start of the next row
          x = 0
          y += 1

        # test if the last row is complete
        if y >= self.env.height:
          # go back to the start
          x = self.env.start[0]
          y = self.env.start[1]

    self.last_start_pos = [x,y]
    return self.last_start_pos


  def single_episode(self):
    '''
      run a single episode to collect the reward for each action of the trajectory
      - virtual base method
    '''
    raise NotImplementedError()


  def get_returns(self):
    raise NotImplementedError()


  def run(self, max_episodes = 1, delta_interval = 10, delta_type = DeltaType.Max, hide_progress = False):

    deltas = []
    for episode in (pbar:=tqdm(range(max_episodes), disable=hide_progress)):

      initial_values = self.returns.copy()

      # set the next start position
      self.env.set_initial_pos(self.get_next_start_state())

      rewards = self.single_episode()
      returns = self.rewards_to_returns(rewards)
      self.get_returns(returns)

      # save the change in the calculated values at regular intervals
      if episode%delta_interval == 0:
        # calculate the difference in the values from the start to end of the iteration
        if delta_type == DeltaType.Max:
          delta = np.max(np.abs(self.returns - initial_values))  # get the largest difference
        else:
          delta = np.mean(np.abs(self.returns - initial_values)) # get the average difference
        deltas.append(delta)
        pbar.set_description(f"Delta {delta:0.5f}")

    return self.returns, self.visits, deltas



class MonteCarloStateValues( MonteCarlo ):
  ''' base class for Monte Carlo methods calculating state values '''

  def __init__(self, policy: Policy, **setup):
    super().__init__(policy, **setup)

    # keep a count of the visits to each state
    self.visits = np.zeros((self.env.height,self.env.width))

    # the average returns for each state
    self.returns = np.zeros((self.env.height,self.env.width))


  def get_graphical_environment(self, text_data):
    ''' create a graphical version of the environment '''
    env = babyrobot.make("BabyRobot-v0", **self.env_setup)
    info = {'text': text_data, 'precision': 0}
    env.show_info(info)
    return env


  def rewards_to_returns(self, state_rewards):
    ''' work backwards to convert the rewards into returns '''
    γ = 1.0
    G = 0
    state_returns = []
    for state, reward in reversed(state_rewards):
      G = reward + γ*G
      state_returns.append((state,G))

    # put back into the order of states visited
    state_returns.reverse()
    return state_returns


  def single_episode(self):
    '''
      run a single episode to collect the reward for each action of the trajectory
    '''
    state,info = self.env.reset()
    state_rewards = []
    total_reward = 0
    terminated = False
    while not terminated:
      # get the policy's action in the current state
      action = self.policy.get_action(self.env.x,self.env.y)

      new_state, reward, terminated, truncated, info = self.env.step(action)
      total_reward += reward
      state_rewards.append((state,reward))
      state = new_state

    return state_rewards


  def get_returns(self, state_returns):
    ''' find the value of the first visit to a state '''
    episode_visits = np.zeros((self.env.height,self.env.width))

    # the first time a state is seen store the return value
    for s, G in state_returns:

      # if first-visit only process the state if it hasn't already been visited
      if self.every_visit or episode_visits[s[1],s[0]] == 0:

        # keep a count of the number of times this state is visited during this episode
        episode_visits[s[1],s[0]] += 1

        # increment the total count of first visits to this state
        self.visits[s[1],s[0]] += 1

        # get (1/visits) setting the result of any divide by zero to zero
        inv_visits = 1 / self.visits[s[1],s[0]]

        # only update the average returns of the states visited in this episode
        self.returns[s[1],s[0]] = ((1 - inv_visits)*self.returns[s[1],s[0]]) + (inv_visits*G)



class MonteCarloActionValues( MonteCarlo ):
  ''' base class for Monte Carlo methods calculating action values '''

  def __init__(self, policy: Policy, epsilon=0, env=None, seed=None, **setup):
    super().__init__(policy, env=env, **setup)

    # set the probability of taking a random action
    self.epsilon = epsilon

    # keep a count of the visits to each action
    self.visits = np.zeros((self.env.height,self.env.width, len(Actions)))

    # the average returns for each action
    self.returns = np.zeros((self.env.height, self.env.width, len(Actions)))

    # set the seed used to choose random actions
    if seed is not None:
      np.random.seed(seed=seed)    


  def get_graphical_environment(self, text_data):
    ''' create a graphical version of the environment '''
    env = babyrobot.make("BabyRobot-v0", **self.env_setup)
    info = {'values': text_data, 'precision': 0}
    env.show_info(info)
    return env


  def rewards_to_returns(self, rewards):
    ''' work backwards to convert the rewards into returns '''
    γ = 1.0
    G = 0
    returns = []
    for state, action, reward in reversed(rewards):
      G = γ*G + reward
      returns.append((state,action,G))

    # put back into the order of actions visited
    returns.reverse()
    return returns


  def single_episode(self):
    '''
      run a single episode to collect the reward for each action of the trajectory
    '''
    state,info = self.env.reset()
    action_rewards = []
    total_reward = 0
    terminated = False
    while not terminated:

      # probability of selecting a random action
      p = np.random.random()

      # if the probability is less than epsilon then a random action
      # is chosen from the state's available actions
      if p < self.epsilon:
        action = self.env.action_space.sample()
      else:
        # get the policy's action in the current state
        action = self.policy.get_action(self.env.x,self.env.y)

      new_state, reward, terminated, truncated, info = self.env.step(action)
      total_reward += reward
      action_rewards.append((state,action,reward))
      state = new_state

    return action_rewards


  def get_returns(self, returns):
    ''' get the return values for each state '''

    # the count of the number of times each action is visited during this episode
    action_visits = np.zeros((self.env.height, self.env.width, len(Actions)))

    # process every state-action pair in the supplied set of returns
    for s, a, G in returns:

      # if first-visit only process the action if it hasn't already been visited
      if self.every_visit or action_visits[s[1],s[0],a] == 0:

        # keep a count of the number of times this action is visited during this episode
        action_visits[s[1],s[0],a] += 1

        # increment the total count of first visits to this state action
        self.visits[s[1],s[0],a] += 1

        # get (1/visits) setting the result of any divide by zero to zero
        inv_visits = 1 / self.visits[s[1],s[0],a]

        # only update the average returns of the state actions visited in this episode
        self.returns[s[1],s[0],a] = ((1 - inv_visits)*self.returns[s[1],s[0],a]) + (inv_visits*G)




class MonteCarloGPI():

  def __init__(self, policy: Policy, evaluation_steps=1, epsilon=0.1, delta_type=DeltaType.Mean, **env_setup):    
    self.policy = policy
    self.evaluation_steps = evaluation_steps
    self.epsilon = epsilon
    self.delta_type = delta_type

    # create an evaluation version of the environment
    self.env = babyrobot.make("BabyRobot-v0", render_mode=None, **env_setup)    

    # keep a count of the first visits to each action
    self.visits = np.zeros((self.env.height, self.env.width, len(Actions)))

    # the average returns for each action
    self.action_values = np.zeros((self.env.height, self.env.width, len(Actions)))    

    # keep track of the deltas over the run
    self.deltas = []


  def do_iteration(self,seed=None):  

        initial_values = self.action_values.copy()     

        mc = MonteCarloActionValues(self.policy, epsilon = self.epsilon, env = self.env, seed = seed)
        eval_values, eval_visits, _ = mc.run(self.evaluation_steps, hide_progress = True)

        # increment the count of any actions that have been visited
        self.visits += eval_visits

        # get (1/visits) setting the result of any divide by zero to zero
        inv_visits = np.reciprocal(self.visits, where=self.visits!=0)

        # only update the average action values of the states visited in this episode
        self.action_values = self.action_values + (inv_visits * (eval_values - (self.action_values*eval_visits)))

        # update the policy with respect to the latest calculated action values
        self.policy.update_policy(self.action_values)  

        # calculate the difference in the action values from the start to end of the iteration
        if self.delta_type == DeltaType.Max:
          delta = np.max(np.abs(self.action_values - initial_values))  # get the largest difference
        else:
          delta = np.mean(np.abs(self.action_values - initial_values)) # get the average difference      

        # save the change in the action values 
        self.deltas.append(delta)          

        return delta      


  def run( self, max_iterations=10, evaluation_steps=1, epsilon=0.1, min_delta=None, delta_interval=1, delta_type=DeltaType.Mean ):
    ''' run 'max_iterations' of policy evaluation and improvement 
      - policy evaluation uses first-visit MC and runs for 'evaluation_steps' steps
    '''      
    
    self.deltas = []
    for episode in (pbar:=tqdm(range(max_iterations))):      

      # perform a single step of MC GPI
      delta = self.do_iteration(evaluation_steps, epsilon, delta_type)
      pbar.set_description(f"Delta {delta:0.5f}")

      # if a minimum delta has been supplied test if the new delta has reached the threshold
      if min_delta is not None and delta < min_delta: 
        break      

    return self.action_values, self.visits, self.deltas, episode



# '''
#     First Visit Monte Carlo
# '''


# class MonteCarlo_FirstVisit( MonteCarloStateValues ):

#   def __init__(self, policy: Policy, **setup):
#     super().__init__(policy, **setup)


#   def get_returns(self, state_returns):
#     ''' find the value of the first visit to a state '''
#     episode_visits = np.zeros((self.env.height,self.env.width))

#     # the first time a state is seen store the return value
#     for s, G in state_returns:
#       # only process the state if it hasn't already been visited
#       if episode_visits[s[1],s[0]] == 0:
#         episode_visits[s[1],s[0]] += 1

#         # increment the total count of first visits to this state
#         self.visits[s[1],s[0]] += 1

#         # get (1/visits) setting the result of any divide by zero to zero
#         inv_visits = 1 / self.visits[s[1],s[0]]

#         # only update the average returns of the states visited in this episode
#         self.returns[s[1],s[0]] = ((1 - inv_visits)*self.returns[s[1],s[0]]) + (inv_visits*G)



# '''
#     Every Visit Monte Carlo
# '''


# class MonteCarlo_EveryVisit( MonteCarloStateValues ):

#   def __init__(self, policy: Policy, **setup):
#     super().__init__(policy, **setup)


#   def get_returns(self, state_returns):
#     ''' use the return from every state visit to calculate the average return '''

#     # the first time a state is seen store the return value
#     for s, G in state_returns:

#       # increment the count of visits to this state
#       self.visits[s[1],s[0]] += 1

#       # get the inverse of visits to this state
#       inv_visits = 1 / self.visits[s[1],s[0]]

#       # only update the average return of the current state
#       self.returns[s[1],s[0]] = ((1 - inv_visits)*self.returns[s[1],s[0]]) + (inv_visits*G)



# '''
#     First-Visit MC Action Values
# '''

# class MonteCarlo_FirstVisit_ActionValues( MonteCarloActionValues ):

#   def __init__(self, policy: Policy, **setup):
#     super().__init__(policy, **setup)


#   def get_returns(self, returns):
#     ''' find the value of the first visit to a state '''

#     episode_visits = np.zeros((self.env.height, self.env.width, len(Actions)))

#     # the first time a state is seen store the return value
#     for s, a, G in returns:
#       # only process the state action if it hasn't already been visited
#       if episode_visits[s[1],s[0],a] == 0:
#         episode_visits[s[1],s[0],a] += 1

#         # increment the total count of first visits to this state action
#         self.visits[s[1],s[0],a] += 1

#         # get (1/visits) setting the result of any divide by zero to zero
#         inv_visits = 1 / self.visits[s[1],s[0],a]

#         # only update the average returns of the state actions visited in this episode
#         self.returns[s[1],s[0],a] = ((1 - inv_visits)*self.returns[s[1],s[0],a]) + (inv_visits*G)